import java.io.BufferedReader;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.PrintStream;
import java.util.ArrayList;
import java.util.List;

import org.apache.log4j.Logger;

import com.trilead.ssh2.Connection;
import com.trilead.ssh2.Session;
import com.trilead.ssh2.StreamGobbler;

/**
 * Esegue un comando su host remoti tramite ssh.
 * 
 * Legge gli host sul quale lanciare il comando dal file hosts.conf presente nella stessa directory.
 * Consultare la documentazione della classe Host per la sintassi richiesta.
 * Viene utilizzata, come metodo di autenticazione di ssh, quello a chiave pubblica.
 * E' richiesta, perciò, la chiave privata avente il seguente path "~/.ssh/id_rsa" (~ e' la home directory)
 * */
public class SSHRemoteLauncher
{
	//stringa da eseguire in remoto
	//public final static String COMMAND = "java -cp log4j-1.2.15.jar:muskel.jar muskel.RemoteInterpreter";
	
	private static final String HOSTS_FILE = "hosts.conf";
	private static final String LOG_FILE_DIR = "./log/";
	
	private static Logger log = Logger.getLogger(SSHRemoteLauncher.class.getName());
	
	
	public static void main(String[] args)
	{
		boolean stdoutLog = false;
		String command = null; //il comando da eseguire in remoto
				
		try{
			command = args[0];
			
			if(command.equals("-s")) {
				printUsage();
				return;
			}		
		}
		catch(IndexOutOfBoundsException e){//parametro necessario non specificato, termino
			printUsage();
			return;
		}
		
		//controllo il parametro opzionale per la stampa su stdout
		try{	
			if(args[1].equals("-s"))
				stdoutLog = true;
			else {
				printUsage();
				return;
			}		
		}catch(IndexOutOfBoundsException e){}//parametro opzionale non specificato, proseguo
		
		//carico la lista degli host dal file
		List<Host> list = null;

		try {
			list = ConfigReader.loadConfig(HOSTS_FILE);
			log.debug("File degli host letto.");
			
			if(list.size() == 0) {
				System.out.println("File degli host vuoto.");
				return;
			}

		} catch (FileNotFoundException e) {
			log.fatal("File degli host non trovato");
			
			return;//non posso continuare, esco
		}

		String homedir = System.getProperty("user.home");

		File keyfile = new File(homedir + "/.ssh/id_rsa"); // o "~/.ssh/id_dsa"
		//TODO supporto a id_dsa

		List<SSHRemoteExec> esecutivi = new ArrayList<SSHRemoteExec>();
	
		//controllo dove scrivere i log
		//e avvio i thread
		
		if(!stdoutLog)//scrivo su file
		{
			//controllo se esiste la directory log ed eventualmente la creo
			File logDir = new File(LOG_FILE_DIR);
			
			if(!logDir.exists())
				if(logDir.mkdir())
					log.trace("log directory creata");
				else {
					System.out.println("Errore durante la creazione della directory di log");
					log.fatal("Errore durante la creazione della directory di log");
					return;//non posso contiuare, esco
				}
			
			long now = System.currentTimeMillis();

			File logFile;

			for (Host host : list)
			{	
				//costruisco il nome del file
				logFile = new File(LOG_FILE_DIR + getLogFilename(host, now));

				SSHRemoteExec exec = new SSHRemoteExec(host, keyfile, command, logFile);

				esecutivi.add(exec);
				exec.start();
				
				log.debug("Avviato thread per "+ host.getHostname());
			}
		}
		else//scrivo su stout come richiesto
		{
			for (Host host : list) {	
				
				SSHRemoteExec exec = new SSHRemoteExec(host, keyfile, command);
				esecutivi.add(exec);
				exec.start();
				log.debug("Avviato thread per "+ host.getHostname());
			}
		}
		
		//attento la fine dei vari thread in esecuzione
		for (SSHRemoteExec exec : esecutivi) {
			try {
				exec.join();
			}
			catch (InterruptedException e) {
				e.printStackTrace();
			}
		}
	}
	
	/**
	 * Restituisce la stringa contenente il nome del logfile.
	 * 
	 * Il formato e' time_hostname:porta_.log
	 * 
	 * @param host l'host da loggare
	 * @param time un istante di tempo in formato java time utile per la differenziazione dei nomefile
	 * */
	private static String getLogFilename(Host host, long time) {		
		
		return time +"_"+ host.getHostname() + ":" + host.getSshPort()+"_.log";		
	}
	
	private static void printUsage()
	{
		System.out.println("Usage: java SSHRemoteLauncher \"command\" [-s]");
		System.out.println("\t-s Prints logs to stdout\n");		
	}
	
}

/**
 * Thread che esegue il comando specificato via ssh.
 * I dati di accesso alla shell remota sono specificati nel file Host.
 * Utilizza per l'autenticazione il metodo a chiave pubblica;
 * tale chiave e' specificata da keyfile.
 * 
 * Stampa lo stdout e lo stderr del processo remoto nel File specificato da logFile (nello stdout locale altrimenti)
 * */
class SSHRemoteExec extends Thread
{
	
	private static Logger log = Logger.getLogger(SSHRemoteExec.class.getName());
	
	private Host host;
	private String command;
	private File keyfile;
	private File logfile;

	/**
	 * Crea un nuovo SSHRemoteExec.
	 * L'output del processo remoto verra' stampato nel file logFile specificato
	 * 
	 * @param host L'oggetto contenente i dati di accesso alla shell remota
	 * @param keyfile Il File della chiave privata necessaria all'autenticazione
	 * @param command Il comando da eseguire in remoto
	 * @param logFile Il File dove scrivere l'output del processo remoto
	 * */
	public SSHRemoteExec(Host host, File keyfile, String command, File logFile) {
		this.host = host;
		this.command = command;
		this.keyfile = keyfile;
		this.logfile = logFile;
	}

	/**
	 * Crea un nuovo SSHRemoteExec.
	 * L'output del processo remoto verra' stampato sullo stdout.
	 * 
	 * @param host L'oggetto contenente i dati di accesso alla shell remota
	 * @param keyfile Il File della chiave privata necessaria all'autenticazione
	 * @param command Il comando da eseguire in remoto
	 * */
	public SSHRemoteExec(Host host, File keyfile, String command) {
		this.host = host;
		this.command = command;
		this.keyfile = keyfile;
		this.logfile = null;
	}

	public void run() {
		try
		{
			PrintStream logStream;

			//se non e' stato specificato il file di log stampo sullo stdout
			if (logfile == null)
				logStream = System.out;
			else
				logStream = new PrintStream(logfile);

			/* Create a connection instance */
			Connection conn = new Connection(host.getHostname(), host.getSshPort());

			/* Now connect */
			conn.connect();
			
			log.debug(host.getHostname()+ ": connesso");

			/* Authenticate */
			boolean isAuthenticated = conn.authenticateWithPublicKey(host.getUsername(), keyfile, host.getPassword());

			if (isAuthenticated == false)
				throw new IOException("Authentication failed.");

			log.debug(host.getHostname()+ ": autenticazione effettuata");
			
			/* Create a session */
			final Session sess = conn.openSession();

			sess.execCommand(command);

			System.out.println("Execution started on " + host.getHostname()+":"+host.getSshPort());
			
			//imposto i thread per leggere dallo sdout...
			InputStream stdout = new StreamGobbler(sess.getStdout());
			BufferedReader br = new BufferedReader(new InputStreamReader(stdout));
			StreamConsumer stdoutConsumer = new StreamConsumer(br, logStream);		

			//... e stderr del processo remoto
			InputStream stderr = new StreamGobbler(sess.getStderr());
			BufferedReader brErr = new BufferedReader(new InputStreamReader(stderr));
			StreamConsumer stderrConsumer = new StreamConsumer(brErr, logStream);

			//avvio i thread
			stdoutConsumer.start();
			stderrConsumer.start();

			//attendo la chiusura dei thread
			try {
				stdoutConsumer.join();				
				stderrConsumer.join();
				
			} catch (InterruptedException e) {
				e.printStackTrace();
			}
			finally{
				/* Close this session */
				sess.close();

				/* Close the connection */
				conn.close();
				log.debug(host.getHostname()+": Connessione chiusa");
			}

			System.out.println("Execution on "+host.getHostname()+" ended");
		}
		catch (IOException e)
		{
			//e.printStackTrace(System.err);
			System.out.println("A problem occurred on "+host.getHostname()+": "+e.getMessage());
			
			//log.fatal("A problem occurred on "+host.getHostname()+": "+e.getMessage());			
		}
	}
	
	
	/**
	 * Thread che consuma un BufferedReader e lo stampa nello stream specificato (di default stdout)
	 * */
	class StreamConsumer extends Thread
	{
		BufferedReader inStream;

		PrintStream outStream;

		/**
		 * Crea un nuovo StreamConsumer con uno specifico outputStream
		 * @param inStream il BufferedReader da consumare
		 * @param outStream Lo stream dove scrivere i dati consumati
		 * */
		public StreamConsumer(BufferedReader inStream, PrintStream outStream) {
			this.inStream = inStream;
			this.outStream = outStream;
		}

		/**
		 * Crea un nuovo StreamConsumer con outputStream di default (System.out)
		 * @param inStream il BufferedReader da consumare
		 * */
		public StreamConsumer(BufferedReader inStream) {
			this.inStream = inStream;
			this.outStream = System.out;
		}

		public void run()
		{	
			String line;
			try
			{
				try
				{
					while ( ( line = inStream.readLine() ) != null )
						outStream.println( line );					
				}
				catch ( IOException e ) {}
				
				inStream.close();
				
				if(outStream != System.out)
					outStream.close();
			}
			catch ( IOException e )	{
				System.err.println("Problem reading the stream" + e.getMessage());
			}
		}
	}
}
