package map;

import java.io.CharArrayWriter;
import java.io.PrintWriter;
import java.lang.reflect.InvocationTargetException;
import java.util.Arrays;

import javax.tools.*;
import javax.tools.JavaCompiler.CompilationTask;

import muskel.Manager;
import muskel.util.JavaSourceFromString;

import skeleton.Compute;
import skeleton.ParCompute;

/**
 * Classe che permette la creazione di Map skeleton
 * con worker e grado di parallelismo specificati.
 * @author albanese
 * */
public class MapFactory {
	
	/**
	 * Restituisce un Map skeleton con worker e grado di
	 * parallelismo specificati
	 * @return Un ParCompute che implementa lo skeleton Map,
	 * 			oppure null se non e' stato possibile crearlo
	 * */
	@SuppressWarnings("unchecked")
	public static synchronized ParCompute newMap(Manager manager, Class<? extends Compute> worker, int degree)
	{
		String className = "Map"+degree;
				
		ParCompute map = null;
		
		//Map e' un ParCompute
		Class<? extends ParCompute> mapClass = null;
		
		boolean classFileAvailable = false;
		
		classFileAvailable = generateCompileMap(worker, degree, className);
				
		if (classFileAvailable) {//il class e' stato compilato
			try {
				//riprovo a caricare la classe
				mapClass = (Class<? extends ParCompute>) Class.forName(className);
			} catch (ClassNotFoundException e) {
				//e.printStackTrace();
				return null;//loading fallito restituisco null			
			}
		}
		else {//la compilazione e' fallita restituisco null
			return null;
		}
			
		try {
			//creo un nuovo oggetto ParCompute usando il primo
			//e unico costruttore di Map che ha manager come parametro
			map = (ParCompute) mapClass.getConstructors()[0].newInstance(manager);
		} catch (IllegalArgumentException e) {		
			e.printStackTrace();
		} catch (SecurityException e) {
			e.printStackTrace();
		} catch (InstantiationException e) {
			e.printStackTrace();
		} catch (IllegalAccessException e) {
			e.printStackTrace();
		} catch (InvocationTargetException e) {
			e.printStackTrace();
		}
		
		return (map != null) ? map : null;
	}

	/**
	 * Genera i sorgenti di una Map con worker, grado di parallelismo
	 * e il nome della classe, specificati. 
	 * Successivamente compila i sorgenti creando nella directory
	 * corrente il file class in modo da poter creare
	 * dinamicamente sue istanze.
	 * @return true se la compilazione non fa generato errori, false altrimenti
	 * */
	private static synchronized boolean generateCompileMap(Class<? extends Compute> worker, int degree, String className) {
		//ottengo il compilatore di sistema
		JavaCompiler compiler = ToolProvider.getSystemJavaCompiler();
		
		//ottengo l'implementazione del file manager standard. Non aggiungo nessun Diagnostic Listener
		StandardJavaFileManager fileManager = compiler.getStandardFileManager(null, null, null);
		
		//Creo il sorgente per il Map
		CreateMapSrc src = new CreateMapSrc(worker, degree);
		
		JavaSourceFromString javaFileObject = null;
		
		//creo un JavaFileObject con la stringa creata da CreateMapSrc
		javaFileObject = new JavaSourceFromString(className, src.getJavaSrc());
		
		//creo la lista di JavaFileObject da compilare
		Iterable<? extends JavaFileObject> fileObjects = Arrays.asList(javaFileObject);
		
		//creo il compilationTask senza nessun listener per gli errori
		CompilationTask task = compiler.getTask(null, fileManager, null, null, null, fileObjects);
		
		//avvio la compilazione
		boolean result = task.call();
		
		return result;
	}
	
}

/**
 * Classe CreateMapSrc, crea i sorgenti dell'implementazione con grafi MDF 
 * di un Map Skeleton con grado di parallelismo e worker specificati.
 * 
 */
class CreateMapSrc {
	/** il nome della classe worker*/
	private String worker;
	/** per indentare */
	private static final String INDENT = "\t";
	/** per indentare */
	private static final String INDENT2 = "\t\t";
	/** il nome della classe da creare*/
	private String className;//Formato: Mapdegree

	private int degree;
	private CharArrayWriter buffer;

	/** PrintBuffer per usufruire della formattazione */
	private PrintWriter pBuffer;

	/**
	 * Costruttore di un' istanza di ParseJava
	 * @param worker
	 *            nome della classe worker (Diverso da <code>null</code>)
	 */
	public CreateMapSrc(Class<? extends Compute> worker, int degree) {

		if (worker == null || degree < 1)
			return;
		
		this.worker = worker.getName();
		this.degree = degree;

		className = "Map"+degree;
		
		// creo gli stream
		buffer = new CharArrayWriter();
		pBuffer = new PrintWriter(buffer);

		printImports();
		pBuffer.println();
		
		printClassDefinition();
		pBuffer.println();
		
		printConstructor();
		pBuffer.println();
		
		printConstructorStaticContent();

		printIdGeneration();
		
		pBuffer.println(INDENT2 + "/*Emitter MDFi setup*/");
		printEmitterDestinations();
		printEmitterMdf();
		
		pBuffer.println();
		
		for (int i = 0; i<degree; i++){
			pBuffer.println(INDENT2 + "/*Worker MDFi n."+i+" setup*/");
			
			printWorkerDestination(i);
			printWorkerMdf(i);
			
			pBuffer.println();
		}
		
		pBuffer.println(INDENT2 + "/*Collector MDFi setup*/");
		printCollectorDestination();
		printCollectorMdf();
		
		pBuffer.println();
		
		printGraphSection();
		
		pBuffer.println();
		
		printClosingBrackets();
		
		pBuffer.close();
	}

	private void printImports() {
		pBuffer.println("import muskel.*;");
		pBuffer.println("import muskel.mdf.*;");
		pBuffer.println("import skeleton.*;");
		pBuffer.println("import examples.*;");
		pBuffer.println("import map.*;");
		pBuffer.println("import static muskel.mdf.Mdfi.*;");
	}
	
	private void printClassDefinition() {
		pBuffer.println("public class "+ className +" extends ParCompute {");
	}
	
	private void printConstructor() {
		pBuffer.println(INDENT + "public " + className + "(Manager manager) {");
	}
	
	private void printConstructorStaticContent() {

		pBuffer.println(INDENT2 + "super(null);");
		pBuffer.println();
		pBuffer.println(INDENT2 + "/* please use your favourite id generator */");
		pBuffer.println(INDENT2 + "IdGenerator idGen = new IntegerIdGenerator();");
		pBuffer.println();
	}
	
	private void printIdGeneration() {
			
		pBuffer.println(INDENT2 + "InstructionId emitter_Id = new InstructionId(idGen.generateId());");
		
		for(int i = 0; i < degree; i++) {
			
			pBuffer.println(INDENT2 + "InstructionId w"+i+"_Id = new InstructionId(idGen.generateId());");
		}
		
		pBuffer.println(INDENT2 + "InstructionId collector_Id = new InstructionId(idGen.generateId());");
		
		pBuffer.println();
		
	}
	
	private void printEmitterDestinations() {

		pBuffer.println(INDENT2 + "Destination[] emitter_dests = new Destination["+ degree + "];");
		pBuffer.println();
		
		String destName;
		
		for (int i = 0; i < degree; i++) {
			
			destName = "d"+ i +"_emitter";
			
			pBuffer.print(INDENT2 + "Destination "+destName+" = new Destination(");
			pBuffer.print("new InputPositionId(new IntegerIdentifier(0)),");
			pBuffer.println(" w"+ i +"_Id, NoGraphId);");
			pBuffer.println(INDENT2 + "emitter_dests["+i+"] = "+ destName +";");
			pBuffer.println();
		}
		
	}
	
	private void printEmitterMdf() {
		
		pBuffer.print(INDENT2 +"Mdfi emitter = new Mdfi(manager, emitter_Id, new MapEmitter(), ");
		pBuffer.println("1, "+ degree +", emitter_dests);");
		
	}
	
	private void printWorkerDestination(int i) {
		
		String destsName = "dests_w" + i;

		String destName = "d_w"+i;
		
		pBuffer.print(INDENT2 + "Destination "+destName+" = new Destination(");
		pBuffer.print("new InputPositionId(new IntegerIdentifier("+i+")),");
		pBuffer.println(" collector_Id, NoGraphId);");
		
		pBuffer.println(INDENT2 + "Destination[] "+destsName+" = {"+destName+"};");
		pBuffer.println();
		
	}
	
	private void printWorkerMdf(int i) {

		pBuffer.print(INDENT2 +"Mdfi w"+ i +" = new Mdfi(manager, w"+ i +"_Id, new "+worker+"(), ");
		pBuffer.println("1, 1, dests_w"+ i +");");
		
	}
	
	private void printCollectorDestination() {
				
		pBuffer.print(INDENT2 + "Destination d0_collector = new Destination(");
		pBuffer.print("new InputPositionId(new IntegerIdentifier(0)), NoInstrId, NoGraphId);");
		pBuffer.println();
		
		pBuffer.println(INDENT2 + "Destination[] collector_dests = {d0_collector};");
		pBuffer.println();
		
	}
	
	private void printCollectorMdf() {

		pBuffer.print(INDENT2 +"Mdfi collector = new Mdfi(manager, collector_Id, new MapCollector(), ");
		pBuffer.println(degree +", 1, collector_dests);");
		
	}
	
	private void printGraphSection() {
		
		pBuffer.println(INDENT2 + "/*MDF graph setup*/");
		pBuffer.println(INDENT2 + "program = new MdfGraph();");
		pBuffer.println();
		
		pBuffer.println(INDENT2 + "program.setInputInstruction(emitter);");
		pBuffer.println(INDENT2 + "program.addInstruction(emitter);");
		
		for (int i = 0; i< degree; i++)
			pBuffer.println(INDENT2 + "program.addInstruction(w"+ i +");");	
		
		pBuffer.println(INDENT2 + "program.addInstruction(collector);");
		pBuffer.println(INDENT2 + "program.setOutputInstruction(collector);");
	}

	private void printClosingBrackets()	{
		pBuffer.println(INDENT+"}\n\n}\n");
	}
	
	/**
	 * Restituisce il sorgente java generato
	 * 
	 * @return il sorgente java generato
	 */
	public String getJavaSrc() {
		return (buffer != null) ? buffer.toString() : "";
	}
	
}