/* This is the muskel: a skeleton/RMI based, parallel programming library
Copyright (C) 2006  Marco Danelutto, Dept. Computer Science, Univ. Pisa, Italy

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA. */

package muskel;

import java.io.Serializable;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Map;

import skeleton.Compute;

import muskel.mdf.IdGenerator;
import muskel.mdf.IntegerIdGenerator;
import muskel.mdf.OperationId;

/**
 * This is used to store the Compute in such a way 
 * that can be later accessed by a OperationId.
 * Only The latter are passed with MDFi,
 * rather than serialized code (which can be huge).
 * The code is serialized just once, at the very
 * beginning of the computation
 * 
 * @author marcod
 * @author albanese
 *
 */
public class CodeStorage implements Serializable {

	//private final boolean debug = Manager.debug;
		
	private static final long serialVersionUID = 8537789175206285538L;

	HashMap<OperationId, Compute> storage = null;
	
	/**
	 * Contiene il class di ogni Compute presente.
	 * Necessario per distinguerli ed evitare duplicati
	 * */
	HashSet< Class<Compute> > storedClasses = null;
	
	IdGenerator idGen = null;
	
	public CodeStorage() {
		
		storage = new HashMap<OperationId, Compute>();
		
		storedClasses = new HashSet<Class<Compute>>();
		
		idGen = new IntegerIdGenerator();
	}
	
	/**
	 * Aggiunge un Compute al CodeStorage, se non già presente,
	 * e ne restituisce l'identificatore.
	 * Qualora il Compute fosse gia' presente,
	 * ne restituisce l'identificatore corrente.
	 * */
	@SuppressWarnings("unchecked")
	public OperationId store(Compute op) {
		
		OperationId id = null;

		Class<Compute> opClass = (Class<Compute>)  op.getClass();

		if(!storedClasses.contains(opClass)) {// nuovo opcode, lo aggiungo

			id = new OperationId(idGen.generateId());

			storedClasses.add(opClass);
			storage.put(id, op);
		}
		else { // opcode esistente cerco l'id corrispondente

			id = getOperationId(op);//ritorna null se op non e' presente, ma ne e' gia' stata controllata la presenza
		}
		
		return id;
	}
	
	public Compute getCompute(OperationId id) throws OpCodeNotPresentException {
		
		Compute toReturn = storage.get(id);
		
		if(toReturn == null)
			throw new OpCodeNotPresentException();
		else return toReturn;
	}
	
	public String toString() {
		String out = "len:"+storage.size()+":\n\t";
		
		Collection<Map.Entry<OperationId, Compute>> entrySet = storage.entrySet();
		
		for (Map.Entry<OperationId, Compute> entry : entrySet) {
			
			OperationId id = entry.getKey();
			
			out = out + ":" + id + ":" + storage.get(id) + ":\n\t";
		}
		
		return out;
	}
	
	
	/**
	 * Restituisce l'identificatore corrispondente
	 * al'opcode <code>comp</code> se l'associazione esiste,
	 * <code>null</code> altrimenti.
	 * @return L'identificatore corrispondente al'opcode
	 * <code>comp</code> se l'associazione esiste,
	 * <code>null</code> altrimenti.
	 * */
	@SuppressWarnings("unchecked")
	private OperationId getOperationId(Compute comp)
	{		
		OperationId id = null;
		
		Class<Compute> compClass = (Class<Compute>) comp.getClass();
		
		Collection<Map.Entry<OperationId, Compute>> entrySet = storage.entrySet();
		
		boolean done = false;
		
		/* scorro le entry per cercare un Compute con Class
		 * uguale a quello del parametro e ne prendo il corrispettivo
		 * OperationId (la key della entry)
		*/
		for (Iterator<Map.Entry<OperationId, Compute>> iterator =  entrySet.iterator();
				!done && iterator.hasNext();)
		{
			
			Map.Entry<OperationId, Compute> entry =
									(Map.Entry<OperationId, Compute>) iterator.next();
			
			Compute opcode = entry.getValue();
			
			Class<Compute> entryClass = (Class<Compute>) opcode.getClass();

			if(entryClass.equals(compClass)){ //stesso Class
				
				id = entry.getKey();
				done = true;//id trovato, ho finito
			}
			
		}
				
		return id;
	}
		
}
