/* This is the muskel: a skeleton/RMI based, parallel programming library
Copyright (C) 2006  Marco Danelutto, Dept. Computer Science, Univ. Pisa, Italy

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA. */

package muskel;


import java.util.Collection;

import muskel.mdf.Destination;
import muskel.mdf.IdGenerator;
import muskel.mdf.InputPositionId;
import muskel.mdf.InstructionId;
import muskel.mdf.IntegerIdGenerator;
import muskel.mdf.IntegerIdentifier;
import muskel.mdf.MdfGraph;
import muskel.mdf.Mdfi;
import muskel.mdf.OperationId;

import org.apache.log4j.Logger;

import skeleton.Compute;
import skeleton.Farm;
import skeleton.ParCompute;
import skeleton.Pipeline;
import skeleton.SECompute;
/**
 * 
 * This is to compile skeletons to MDF graphs
 * @author marcod
 *
 */
public class Compiler {

    private static Logger log = Logger.getLogger(Compiler.class.getName());
    
    private static IdGenerator idGen = new IntegerIdGenerator();

	public static MdfGraph compile(Compute prgm, CodeStorage cs) {
		log.debug("Compiling => "+(new PrettyPrint(prgm)).toString());
		if( prgm instanceof Farm) {
			// this is the trivial case, just compile the workers
			Compute worker =  ((Farm) prgm).getWorker();
			MdfGraph w = compile(worker,cs);
			log.debug("Compiled ==> \n"+w.toString());
			log.debug("<==");
			return w;
		}
		if(prgm instanceof Pipeline) {
			// this is the composite case
			// get the two graphs
			Compute firstStage = ((Pipeline) prgm).getFirstStage();
			Compute secondStage = ((Pipeline) prgm).getSecondStage();
			MdfGraph firstGraph = compile(firstStage,cs);
			MdfGraph secondGraph = compile(secondStage,cs);
						
			// link the two
			InstructionId secondGrFirstInstrId = secondGraph.getInputInstrId();
			InstructionId firstGrFirstInstrId = firstGraph.getInputInstrId();
			// ATTENTION: this works only on 1 in 1 out instructions !!! due to the 0 first parameter
			
			// this links the two
			Destination newdest = new Destination(new InputPositionId(new IntegerIdentifier(0)), secondGrFirstInstrId, Mdfi.NoGraphId);
			
			log.debug("First stage graph is :: "+firstGraph.toString());
			log.debug("Dest is ::"+newdest);
			log.debug("Setting dest in graph "+firstGraph.toString());
			
			InstructionId last = firstGraph.setOutputInstrId(newdest); // should be void ... returns the instrID changed (the last)...
			log.debug("Dest id was "+last+" graph is now :: "+firstGraph.toString());
			
			// now return the new graph: merging the first one and the relocated second one
			MdfGraph newgraph = new MdfGraph(Mdfi.NoGraphId);
			
			//la prima istruzione del primo grafo
			Mdfi firstGrFirstInstr = firstGraph.getGraph().get(firstGrFirstInstrId);
			//l'ultima istruzione del secondo grafo
			Mdfi secondGrLastInstr = secondGraph.getGraph().get(secondGraph.getOutputInstrId());
			
			//imposto la prima istruzione del primo grafo come istruzione di input del nuovo grafo, 
			newgraph.setInputInstruction(firstGrFirstInstr);	
			
			Collection<Mdfi> iterableFirstGraph = firstGraph.getGraphMdfi();
			Collection<Mdfi> iterableSecondGraph = secondGraph.getGraphMdfi();
			
			//aggiungo al nuovo grafo le istruzioni del primo
			for (Mdfi mdfi : iterableFirstGraph) 
					newgraph.addInstruction(mdfi);
			
			//aggiungo al nuovo grafo le istruzioni del secondo
			for (Mdfi mdfi : iterableSecondGraph) 
				newgraph.addInstruction(mdfi);
			
			//imposto l'ultima istruzione del secondo grafo come istruzione di output del nuovo grafo, 
			newgraph.setOutputInstruction(secondGrLastInstr);	

			log.debug("Compiled ==> \n"+newgraph.toString());
			log.debug("<==");
			return newgraph;
		}
		if(prgm instanceof ParCompute) {
		    // this is the case the graph is already provided by the user in the body of the ParCompute object
		    // actually this is the simplest of the cases handled here:
			
		    MdfGraph g =  (((ParCompute) prgm).getMdfGraph()).mdfClone();
		    log.debug("got ParCompute graph "+g);
		    return g;
		}

		// no other case than a seq code Compute: this is an Object to Object ... must be a Object[] to Object[]
		MdfGraph seqgraph = new MdfGraph();
		
		OperationId opcode = cs.store(prgm);

		Destination[] dest1 = {new Destination(new InputPositionId(new IntegerIdentifier(0)), Mdfi.NoInstrId, Mdfi.NoGraphId)};
		
		//genero un nuovo id per la nuova mdfi
		InstructionId seqmdfiId = new InstructionId(idGen.generateId());
		
		boolean seCompute = (prgm instanceof SECompute);
		
		Mdfi seqmdfi = new Mdfi(seqmdfiId, Mdfi.NoGraphId, opcode, 1, 1, dest1, seCompute);
		
		seqgraph.setInputInstruction(seqmdfi);
		seqgraph.addInstruction(seqmdfi);
		seqgraph.setOutputInstruction(seqmdfi);
		
		log.debug("Compiled ==> \n"+seqgraph.toString());
		log.debug("<==");
		return seqgraph;
	}
}
