/* This is the muskel: a skeleton/RMI based, parallel programming library
Copyright (C) 2006  Marco Danelutto, Dept. Computer Science, Univ. Pisa, Italy

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA. */

package muskel;

import java.net.InetAddress;
import java.net.MalformedURLException;
import java.net.UnknownHostException;
import java.rmi.Naming;
import java.rmi.NotBoundException;
import java.rmi.RemoteException;


import muskel.mdf.*;
import muskel.stream.*;

import org.apache.log4j.Logger;

/**
 * Used to check the compiler/storage/pool
 * 
 * @author marcod
 * 
 */
public class ControlThread implements Runnable {

	public final boolean debug = Manager.debug;

	private CodeStorage cs = null;

	private MdfiPool pool = null;

	private OutputManager osm = null;
	private InputManager ism = null;

	private String host = "";
	
	private boolean blockingInStream = false;
	
	public ControlThread(CodeStorage ccs, MdfiPool ppool, InputManager iism, OutputManager oosm,
			String hhost) {
		cs = ccs;
		pool = ppool;
		osm = oosm;
		host = hhost;
		ism = iism;
		
		/*prendo nota se lo stream e' bloccante e permette
		 * la produzione dei task da parte dei Compute
		 */
		blockingInStream = (ism instanceof BlockingInputManager);
		
	}

	public void run() {
		// set up the remote node connection
		RemoteInterpreterInterface worker = null;
		
		Logger log = Logger.getLogger(ControlThread.class.getName());
		
		log.info("Control thread started");
		try {
			InetAddress ciccio = InetAddress.getByName(host);

			log.debug("Got address " + ciccio + " for " + host);
			String stringa = "rmi://" + host + ":" + RemoteInterpreter.Port
					+ "/muSkelMDFWorker";
			if (debug)
				System.out.println("Looking up worker on host " + stringa);
			
			worker = (RemoteInterpreterInterface) Naming.lookup(stringa);
			
			log.debug("Remote worker looked'up");
			if (debug)
				System.out.println("worker is " + worker);
		} catch (MalformedURLException e) {
			e.printStackTrace();
		} catch (RemoteException e) {
			e.printStackTrace();
		} catch (NotBoundException e) {
			e.printStackTrace();
		} catch (UnknownHostException e) {
			e.printStackTrace();
		}

		if (worker == null) {
			//TODO non deve uscire, ma cercare un altro worker (da implementare)
			System.out.println("Cannot find worker");
			return;			
		}
		log.info("RemoteInterpeter found");

		log.debug("Staging program to remote intepreter... ");
		
		try {
			worker.setProgram(cs);
		} catch (RemoteException e) {
			log.fatal("setProgram failed on remote node (Exception is " + e
					+ " )");
			e.printStackTrace();
			return;
		}
		
		log.info("code staged");

		log.debug("Control thread " + this
				+ " getting fireable instruction ... ");
		
		Mdfi instr = pool.getFireable(); // get an instruction to compute; in
											// case it blocks
		
		if (instr == null)
			return; // terminate
		
		log.debug("Control thread " + this + " fetched fireable "
				+ instr.toString());
		
		log.info("Entering main loop");
		
		do {
			ComputeResult res = null;
			try { // compute the instruction remotely
				res = worker.compute(instr);
			} catch (RemoteException e) {
				// TODO this is where we can trigger the manager
				// first: put back the not computed mdf instruction
				pool.insert(instr);
				
				// then inform the manager TODO
				// eventually terminate the thread
				log.fatal("remote worker terminated due RemoteException on mdfi compute request");
				
				//se lo stream supporta i produttori e se l'instr e' un produttore
				if(blockingInStream && instr.isSeMDFi())
					//segnalo allo stream la perdita di un produttore
					((BlockingInputManager) ism).removeProducer();
					
				e.printStackTrace();
				return; // this terminates the thread actually
			}
			
			/* se l'input stream supporta l'aggiunta di task e la compute
			 * calcolata ha side effects aggiungo i task calcolati
			 * allo stream ed gli segnalo che un produttore ha terminato
			 * */
			 
			if(blockingInStream && instr.isSeMDFi()) {
				
				BlockingInputManager bism = (BlockingInputManager) ism;
				SEComputeResult seRes = (SEComputeResult) res;
				
				Task[] genTasks = seRes.getGeneratedTasks();
				
				if (genTasks != null) {
					for (Task task: genTasks)
						bism.addTask(task.getValue());
				}		
				bism.removeProducer();
			}
			
			//gestisco i risultati, caso comune sia a Compute che a SECompute
			OutputToken[] outTokens = res.getOutTokens();
			
			Destination dest;
			for (OutputToken outTok: outTokens) {// deliver the results
				log.info("Result token : " + outTok);

				dest = outTok.getDestination();
				
				if (dest.getInstructionId().equals(Mdfi.NoInstrId)) {//controllo se e' l'ultimo token del grafo
					try {
						osm.deliver(outTok.getTask());//lo spedisco all'output stream
					} catch (TokenNotPresentException e) {}//il token non ha un task
					
				} else {
					//inserisco il token nel pool
					log.info("Going to store token " +outTok);
					pool.putToken(outTok);
					log.info("Control thread " + this + " stored token "
							+ outTok.toString());
				}
			}
						
			// now get another instruction to compute
			instr = pool.getFireable(); // get an instruction to compute; in
										// case it blocks
			log.debug("Control thread " + this + " fetched fireable " + instr);
			if (instr == null) { // this is the case the thread gets an
									// INTERRUPT from the manager, that verified
									// all threads are blocked on
				// the empty mdfi pool getFireable ...
				// therefore statistics are gathered from the associated worker,
				// they are printed and then the thread terminates
				log.debug("Going to retrieve the remote stats ...");
				try {
					Stats s = worker.getStats();
					log.debug("Retrieved");
					log.info("Stats: " + s);
					System.out.println(s);
				} catch (RemoteException e) {
					e.printStackTrace();
				}
				return; // terminate
			}

		} while (true); // TODO termination

	}
}
