/* This is the muskel: a skeleton/RMI based, parallel programming library
Copyright (C) 2006  Marco Danelutto, Dept. Computer Science, Univ. Pisa, Italy

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA. */

/*
 * Created on Mar 26, 2004
 *
 * To change the template for this generated file go to
 * Window - Preferences - Java - Code Generation - Code and Comments
 */
package muskel;

import java.util.Vector;
import org.apache.log4j.Logger;

/**
 * This class is used to maintain a list of available
 * (discovered, actually) processing resources,
 * that is processing elements in the portion of network accessible
 * via multicast messages that currently run a muskel RemoteWorker
 * RMI object. <br>
 * The class uses a DiscoveryThread thread to actually
 * perform remote processing elements discovery.
 * <br>
 * The list of workers maintained by this class 
 * is a simple Vector. New workers requested by the ControlThread 
 * are provided from the start position. <br>
 * New workers discovered are added at the end.
 * <br> 
 * A list of already given worker (to the ControlThread) is also maintained.
 * <br>
 * In case a number nw of processing elements is requested,
 * which is not currently available, the calling thread
 * is actually blocked (wait()) and it is eventually
 * unblocked by the DiscoveryThread communicating 
 * that a new resource is available.
 * 
 * @author Danelutto Marco
 *
 *
 */
public class DiscoveryService {
	
	/** this is the vector hosting the hosts discovered but not yes used */
	Vector<String> discoveredWorkers = null;
	/** this is the vector hosting the hosts discovered and already used */
	Vector<String> givenWorkers = null;
	
	/** this is the log file handle */
	Logger logFile;
	
	/**  
	 * standard constructor
	 * initializes the vectors and starts the discovery thread 
	 */
	public DiscoveryService() {
		if(Manager.debug)
			System.out.println("--> Discovery service creation");
		
		logFile = Logger.getLogger(DiscoveryService.class.getName());
		
		discoveredWorkers = new Vector<String>();
		givenWorkers = new Vector<String>();
		
		DiscoveryThread dt = new DiscoveryThread(this);
		dt.setDaemon(true);
		dt.start();
		
		return;
	}
			
	/**
	 * used from the StandardEval thread to get a fresh
	 * host to run a worker.
	 * @return the name of the fresh host discovered
	 */
	public synchronized String getNewWorker() {
		while(discoveredWorkers.size()==0) {
			try {
				wait();
			} catch (InterruptedException e) {
				e.printStackTrace();
			}
		}
		String worker = (String) discoveredWorkers.remove(0);
		givenWorkers.add(worker);
		logFile.debug("getNewWorker returns "+worker);
		return worker;
	}

	/**
	 * used from within the Standard compute method to get a set of workers 
	 * @param nw the number of workers required
	 * @return the vector with the worker
	 * 	(TODO this actually can be smaller than the requested
	 * 				size in case there are not enough workers)
	 */
	public synchronized String [] getNWorkers(int nw) {
		while(discoveredWorkers.size() < nw) {
			try {
				wait();
			} catch (InterruptedException e) {
				e.printStackTrace();
			}
		}
		String [] v = new String [nw];
		for(int i=0; i<nw; i++) {
			v[i] = (String) discoveredWorkers.remove(0);
			givenWorkers.add(v[i]);
		}
		logFile.debug("getNWorkers returned a vector with "+nw+" workers");
		return v;
	}
	
	/**
	 * Used by the DiscoveryThread to place the name of
	 * a fresh host discovered in the 
	 * discovered hosts vector.
	 * Checks for presence of that host in both
	 * the fresh and the given vector before adding
	 * as the Discovery thread does not hold the state
	 * of already discovered hosts.
	 * 
	 * @param w the name of the host to be added
	 */
	public synchronized void addWorker(String w) {
		if(!(discoveredWorkers.contains(w) || givenWorkers.contains(w))) {
			discoveredWorkers.add(w);
			logFile.debug("addWorker added "+w);
			notifyAll();
		}
		return;
	} 
}
