/* This is the muskel: a skeleton/RMI based, parallel programming library
Copyright (C) 2006  Marco Danelutto, Dept. Computer Science, Univ. Pisa, Italy

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA. */

package muskel;

import java.net.InetAddress;
import java.net.UnknownHostException;
import muskel.mdf.*;
import muskel.stream.BlockingInputManager;
import muskel.stream.InputManager;
import muskel.stream.OutputManager;
import org.apache.log4j.Logger;
import skeleton.Compute;

/**
 * takes care of a program execution request, with contract (#PEs).
 * A manager object should be instantiated to run muskel programs.
 * The manager is to be given the program,
 * the input and output stream and a performance contract.
 * The manager can be asked to "compute" the program
 * by invoking the homonymous method. 
 * 
 * @author marcod
 *
 */
public class Manager {

	public static final boolean debug = false;
	boolean localDebug = false; 

	Compute program = null; 
	CodeStorage cs = null; 
	Contract pc = null; 
	DiscoveryService ds = null; 
	MdfiPool pool = null; 
	InputManager ism = null; 
	OutputManager osm = null; 
	
	InstructionId firstInstructionId = Mdfi.NoInstrId;
	
	private boolean blockingInStream = false;
	
	IdGenerator idGen;
	
	Logger log = null; 
	 // this will host the compiled graph (plus user provided mdfis)
	MdfGraph theGraph = null;

	/** This constructs the manager. 
	 * @param ppgm the program to be computed. It must implement the Compute interface
	 * @param iism the input manager, a class implementing 
	 * 		the InputManager Interface that requires the methods boolean
	 *		hasNext() (telling whether or not the input stream has further items)
	 *		and an Object next() method that returns the next item of the input stream
	 * @param oosm the output manager. Its void deliver(Object res)
	 * 		method is called each time a new result is available
	 */
	public Manager(Compute ppgm, InputManager iism, OutputManager oosm) {
		
		log = Logger.getLogger(Manager.class.getName());
		program = ppgm;
		cs = new CodeStorage();

		idGen = new IntegerIdGenerator();
		
		/*Se il Compute e' disponibile lo compilo.
		 * Altrimenti lo faro' quando sarà reso disponibile tramite la setProgram.
		 * (e' il caso dei ParCompute)
		 */
		if(ppgm != null) { // Il Compute da compilare e' disponibile
			theGraph = Compiler.compile(program,cs);
			log.debug("Compiled graph is "+theGraph);
		}
		
		// set up discovery
		ds = new DiscoveryService();
		
		pool = new MdfiPool();
		ism = iism; 
		osm = oosm;
		
		/*prendo nota se lo stream e' bloccante
		 *e permette la produzione dei task da parte dei Compute
		*/
		blockingInStream = (ism instanceof BlockingInputManager);
	}

	/** this constructs the manager without giving all the parameters.
	 * It is used when user defined MDF graphs are 
	 * supplied in the program via the ParCompute class. 
	 * Later on the missing parameters are set up with the proper accessor methods. 
	 */ 
	public Manager() {
		this(null, null, null); 
	}

	/** Accessor method to set program.
	 * 	Also compile the <code>pgm</code> mdfgraph
	 * @param pgm the program to be computed by the manager
	 */
	public void setProgram(Compute pgm) {
		program = pgm;
		
		theGraph = Compiler.compile(program, cs);
	}

	/** accessor method to set the input manager
	 * @param im the input manager
	 */
	public void setInputManager(InputManager im) {
		ism = im;
	}

	/** accessor method to set the output manager
	 * @param om the output manager
	 */
	public void setOutputManager(OutputManager om) {
		osm = om;
	}
	/** the method is used to set up the performance contract with the manager.
	 * The user asks a given parallelism degree by stating it in the contract. 
	 * AT the moment, the request is satisfied all or nothing.
	 * IF less than the requested number of remote intepreters are 
	 * found, then the manager blocks on the compute method. 
	 * @param c the performance contract to be satisfied by the manager
	 */
	public void setContract(Contract c) {
		pc = c;
	}


	/**
	 * this method returns a pointer to the CodeStorage.
	 * It is used to implement user defined macro data flow instructions. 
	 * User defined Compute code is used as the MDFi opcode.
	 * The Compute code has to be stored into the CodeStorage,
	 * in such a way it can be retrieved (with an integer pointer)
	 * when the MDFi has to be executed.
	 * @return the current CodeStorage pointer
	 */ 
	public CodeStorage getCodeStorage() {
		return cs;
	}


	/** Sets a new Compute code in the CodeStorage,
	 * returning the OperationId in the CodeStorage.
	 * If the CodeStorage doesn't exist, returns Mdfi.NoOpCode
	 * @param opc The Compute program to be stored in the CodeStorage
	 * @return Tthe OperationId in the CodeStorage.
	 * 			Mdfi.NoOpCode if the CodeStorage not exists.
	 */
	public OperationId storeOpCode(Compute opc) {
		
		OperationId opcode;
		
		try {
			opcode = cs.store(opc);
			
		} catch (NullPointerException e) {
			log.fatal("storeOpCode with an empty CodeStorage");
			
			opcode = Mdfi.NoOpCode;
		}
		
		return opcode;
	}


	/**
	 * Gives the programmer the ability to retrieve
	 * the compiled graph of a skeleton program.
	 * This can be used to merge with ad hoc data flow graphs,
	 * provided by the user, that express parallelism exploitation
	 * patterns not provided within the native skeletons.
	 * 
	 * @param pgm The program to be mdf compiled
	 * @return The compiled graph of pgm.
	 */
	public MdfGraph newGraph(Compute pgm) {
		MdfGraph graph = Compiler.compile(pgm,cs);
		return graph;
	}


	/** the method starts the parallel computation of
	 * the program stored in the manager. 
	 * A number of remote processing elements is recruited,
	 * according to the performance contract passed to the manager
	 * then an according number of control threads are started.
	 * Eventually, they cooperate to compute the program on the 
	 * distributed network of MDF interpreters
	 */

	public void compute() {
		if(program==null) {
			log.fatal("computing a null program");
			return;
		}
		if(ism== null) {
			log.fatal("computing program without input manager");
			return;
		}
		if(osm== null) {
			log.fatal("computing program without output manager");
			return;
		}
		
		/*
		 *  set up a proper pool of ControlThreads accessing 
		 *  he MdfiPool and delivering fireable
		 *  MDFis to remote interpreters in a loop
		*/
		long t0 = System.currentTimeMillis();
		log.debug("Manager: start");
		int nw = ((ParDegree) pc).getParDegree();  // look for the contract
		
		String [] machines = ds.getNWorkers(nw);	// search machines
		// may be less than nw TODO
		muskel.ControlThread [] ct  = new muskel.ControlThread[machines.length];	
		
		// actual threads here (to use groups)
		Thread [] controlThreads = new Thread[machines.length];
		// start threads in a thread group ... to control termination with active count
		ThreadGroup thGr = new ThreadGroup("controlThreads");

		// for each machine start a control thread
		for(int i=0; i<machines.length; i++) {
			try {
				ct[i] = new muskel.ControlThread(cs, pool, ism, osm, (InetAddress.getByName(machines[i])).getHostName());
			} catch (UnknownHostException e) {
				e.printStackTrace();
			}
			controlThreads[i] = new Thread(thGr,ct[i],"controlThread"+i);
			controlThreads[i].start();
		}
		
		long t1 = System.currentTimeMillis();
		long elapsed = (t1 > t0 ? t1 - t0 : Long.MAX_VALUE - t0 + t1);
		log.info("Total elapsed time (init control threads): "+elapsed+" msec with "+machines.length+" PEs");
		
		log.debug("Manager: thread started");
		
		GraphId newGid;
			
		// now start fetching input tasks and delivering them to the pool with the associated graph
		while(ism.hasNext()) {
			
			// fetch an input task
			
			Task task = new Task(ism.next());
			
			log.debug("Manager: task fecthed "+task);
			
			int numSeCompute = theGraph.getNumSECompute();
			
			/*segnalo allo stream la presenza di nuovi produttori
			 *se lo stream li supporta e se il grafo ha
			 *istruzioni che producono task
			*/
			if(blockingInStream &&  numSeCompute> 0)
				((BlockingInputManager) ism).addProducers(numSeCompute);
			
			//set graphId
			MdfGraph graph = theGraph.mdfClone();//clone a new graph for each token
			log.debug("Cloned graph is "+graph);
			
			firstInstructionId = graph.getInputInstrId(); // this is the entry point
			
			newGid = new GraphId(idGen.generateId());
			
			graph.setGid(newGid);	// instantiate the graph
			
			// add the input token
			OutputToken outToken = new OutputToken(task, new Destination(
									new InputPositionId(new IntegerIdentifier(0)),firstInstructionId, newGid));
										
			pool.insert(graph);// put the graph in the pool
			
			
			log.info(">>>>> "+graph);
			
			pool.putToken(outToken);
			
			log.debug("Manager: graph inserted in pool");
			
		}
		
		log.debug("Manager: task insertion terminated");
		
		long t2 = System.currentTimeMillis();
		elapsed = (t2 > t0 ? t2 - t0 : Long.MAX_VALUE - t0 + t2);
		log.info("Total elapsed time (tasks inserted): "+elapsed+" msec with "+machines.length+" PEs");
		
		// now wait the control threads ...
		log.debug("Thread Group has "+thGr.activeCount()+" active Threads");
		
		int blocked = 0;
		
		while(thGr.activeCount() >= machines.length && blocked != machines.length) {
			
			try {
				Thread.sleep(100);
			} catch (InterruptedException e) {
				e.printStackTrace();
			}  // just for releasing the CPU
			
			log.debug("ACTIVE COUNT "+thGr.activeCount());
			blocked = 0; // added to avoid accumulation ...
			
			for(int i=0;i<machines.length; i++) {
				log.debug(">>> "+controlThreads[i]+" ("+controlThreads[i].getState()+")");
				if(controlThreads[i].getState() == Thread.State.WAITING) 
					blocked++;
			}
			
			if(blocked == machines.length) {
				// terminate threads
				//TODO after gathering statistics ... 
				for(int i=0; i<machines.length; i++)
					controlThreads[i].interrupt();
			}
		}
		
		log.debug("All threads blocked ...");
		
		for(int i=0; i<machines.length; i++) {
			try {
				log.debug("Joining thread "+i);

				controlThreads[i].join();

				log.debug("Thread "+i+" joined");
			} catch (InterruptedException e) {
				e.printStackTrace();
			}
		}
		
		long t3 = System.currentTimeMillis();
		elapsed = (t3 > t0 ? t3 - t0 : Long.MAX_VALUE - t0 + t3);
		System.out.println("Total elapsed time: "+elapsed
								+" msec with "+machines.length+" PEs");
		System.exit(0); // kills also the discovery service thread
	}

}
