/* This is the muskel: a skeleton/RMI based, parallel programming library
Copyright (C) 2006  Marco Danelutto, Dept. Computer Science, Univ. Pisa, Italy

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA. */

package muskel;

import java.util.HashMap;
import java.util.Vector;
import muskel.mdf.*;
import org.apache.log4j.Logger;

/**
 * This is used to store Mdfi from graphs.
 * 
 * @author marcod
 *
 */
public class MdfiPool {

	public final boolean debug = Manager.debug;

	HashMap<InstructionTag,Mdfi> pool = null;

	Vector<InstructionTag> fireableInstructions = null;
	
	Logger log = null; 

	public MdfiPool() {

		pool = new HashMap<InstructionTag,Mdfi>();
		fireableInstructions = new Vector<InstructionTag>();

		log = Logger.getLogger(MdfiPool.class.getName());
	}
	
	/**
	 * Inserts <code>instruction</code> into the pool
	 * */
	public synchronized void insert(Mdfi instr) {
		
		InstructionTag iTag = new InstructionTag(instr.getInstrId(),instr.getGraphId());
		
		pool.put(iTag, instr);

		if(instr.isFireable()) 
			fireableInstructions.add(iTag);
		
		notifyAll();//notifico anche se l'istr non è fireable perchè c'è la extract
	}
	
	/**
	 * Inserts the <code>graph</code> mdfi into the pool
	 * */
	public synchronized void insert(MdfGraph graph) {
		
		InstructionTag iTag;
		
		for(Mdfi instr: graph.getGraphMdfi()) {
			
			iTag = new InstructionTag(instr.getInstrId(),instr.getGraphId());
			
			pool.put(iTag, instr);
			if(instr.isFireable())
				fireableInstructions.add(iTag);
			//notifico anche se l'istr non è fireable perchè c'è la extract
			notifyAll();
		}
	}
	
	/**
	 * Extracts the mdfi matching to <code>instrTag</code> from the pool.
	 * */
	public synchronized Mdfi extract(InstructionTag instrTag) {
		while(pool.isEmpty()) {
			try {
				wait();
			} catch (InterruptedException e) {
				e.printStackTrace();
			}
		}
		
		Mdfi instr = pool.remove(instrTag);
		
		return instr;
	}
	

	/**
	 * Extracts the mdfi having <code>instr</code> and
	 * <code>graph</code> as InstructionId and GraphId.
	 * */
	public synchronized Mdfi extract(InstructionId instr, GraphId graph) {

		return extract(new InstructionTag(instr,graph));
	}
		
	/**
	 * Returns the first fireable mdfi (according to a FIFO policy)
	 * @return Returns the first fireable mdfi
	 * */
	public synchronized Mdfi getFireable() {
		while(fireableInstructions.isEmpty()) {
			try {
				log.warn("getFireable: empty Fireable instruction pool\n"+pool);
				wait();
			} catch (InterruptedException e) { // if you get there, you're terminating
				return null;
			}
		}
		InstructionTag istrTag = fireableInstructions.remove(0); // this is FIFO
		
		Mdfi instr = pool.remove(istrTag);
		log.debug("getFireable: got "+instr);
		
		return instr;
	}	
	
	/**
	 * Inserisce outToken nella rispettiva mdfi (specificata in outToken stesso)
	 * @throws IllegalArgumentException Se outToken non contiene un task
	 * */
	public synchronized void putToken(OutputToken outToken) {
		
		Task task;
		
		try {
			task = outToken.getTask();
		} catch (TokenNotPresentException e) {//il token non contiene un task
			
			throw new IllegalArgumentException("Token uninitialized");
		}
		
		Destination dest = outToken.getDestination();

		InstructionTag instrTag = new InstructionTag(dest.getInstructionId(), dest.getGraphId());
				
		log.debug("Putting input token "+outToken.toString()+" InstrutionId is "+instrTag.getInstrId()+" GraphId is "+instrTag.getGraphId());
		
		//prendo la mdfi dal pool
		Mdfi instr = pool.get(instrTag);
		
		/*inserisco il task di outToken nel vettore di input di instr
		 * (nella posizione specificata in outToken)
		 */ 
		instr.storeToken(dest.getInputPositionId(), task);	
		
		//controllo se l' instr sia diventata fireable con in nuovo token 
		if(instr.isFireable()) {
			fireableInstructions.add(instrTag);
			log.debug("putToken: added "+instrTag);
			
			notifyAll();//notifico i ControlThread in attesa di istruzioni fireable
		}
		
		return;
	}
	
	public String toString() {
		String p = "Pool ("+pool.size()+" instr, "+fireableInstructions.size()+" fireable):\n";
		
		for(Mdfi instr: pool.values()) 
			p = p + instr.toString() + "\n";
		
		p = p + "\n";
		
		for(InstructionTag instrTag :fireableInstructions)		
			p = p + instrTag.toString();
		
		p = p + "\nENDPOOL\n";
		return p;
	}
	
}
