/* This is the muskel: a skeleton/RMI based, parallel programming library
Copyright (C) 2006  Marco Danelutto, Dept. Computer Science, Univ. Pisa, Italy

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA. */

package muskel;

import java.net.InetAddress;
import java.net.UnknownHostException;
import java.net.URLClassLoader;
import java.net.URL;
import java.rmi.AccessException;
import java.rmi.RMISecurityManager;
import java.rmi.Remote;
import java.rmi.RemoteException;
import java.rmi.registry.LocateRegistry;
import java.rmi.registry.Registry;
import java.rmi.server.UnicastRemoteObject;
import muskel.mdf.ComputeResult;
import muskel.mdf.Mdfi;
import org.apache.log4j.Logger;

public class RemoteInterpreter extends UnicastRemoteObject implements
		RemoteInterpreterInterface {

	private static Logger log = null;

	private static final long serialVersionUID = 1L;

	public static final int Port = 12460;

	CodeStorage cs = null;

	// this is for statistics
	int mdfiCount; // the number of MDFi computed

	long timeCompute; // the total time spent computing them

	long maxTimeCompute; // the maximum time spent computing a MDFi

	long minTimeCompute; // the minimum time spent computing a MDFi

	boolean firstTime;

	protected RemoteInterpreter() throws RemoteException {
		super();
		log = Logger.getLogger(RemoteInterpreter.class.getName());
		mdfiCount = 0;
		timeCompute = 0;
		firstTime = true;
		log.info("Created remote interpreter");
		String url = "http://192.168.1.100/marcod/Muskel/";
		// setting up the proper classloader (URL, at the moment)
		ClassLoader previous = Thread.currentThread().getContextClassLoader();
		// Create a class loader using the URL as the codebase
		// Use previous as parent class loader to maintain current visibility
		ClassLoader current = null;
		try {
			current = URLClassLoader.newInstance(new URL[] { new URL(url) },
					previous);
		} catch (java.net.MalformedURLException e) {
			log.warn(e + " while setting classloader in RemoteInterpreter");
			e.printStackTrace();
		}
		Thread.currentThread().setContextClassLoader(current);
		log.info("url classloader appended");
		return;
	}

	// this is the part executing the interpreter
	public void setProgram(CodeStorage ccs) throws RemoteException {
		//TODO rifiutare l'esecuzione se l'interprete e' occupato
		//per evitare la sovrapposizione del nuovo codestorage
		//con quello corrente
		
		try {
			cs = ccs;
			log.info("CodeStorage initiated");
			mdfiCount = 0;
			timeCompute = 0;
			firstTime = true;
		} catch (Exception e) {
			log.warn("Exception " + e + " while loading code");
			e.printStackTrace();
		}
	}

	public Stats getStats() {
		Stats s = new Stats();
		s.setMdfiNo(mdfiCount);
		s.setAverageTaskTc(timeCompute / mdfiCount);
		s.setMinTimeCompute(minTimeCompute);
		s.setMaxTimeCompute(maxTimeCompute);
		try {
			s.setWorkerName(InetAddress.getLocalHost().getHostName());
		} catch (UnknownHostException e) {
			e.printStackTrace();
		}
		return s;
	}

	public ComputeResult compute(Mdfi instr) throws RemoteException {
		if (cs == null) {
			log.error("compute request with null codestorage");
			return null;
		} else {
			ComputeResult result = null;
			try {
				log.info("going to compute:" + instr);
				long t0 = System.currentTimeMillis();

				result = instr.compute(cs);

				long t1 = System.currentTimeMillis();
				long elapsed = (t1 >= t0 ? t1 - t0 : Long.MAX_VALUE - t0 + t1);
				timeCompute += elapsed;
				if (firstTime) {
					maxTimeCompute = elapsed;
					minTimeCompute = elapsed;
					firstTime = false;
				} else {
					if (elapsed > maxTimeCompute)
						maxTimeCompute = elapsed;
					if (elapsed < minTimeCompute)
						minTimeCompute = elapsed;
				}
				mdfiCount++;
			} catch (NonFireableInstructionException e) {
				e.printStackTrace();
				result = null; // TODO cosi' non si vedono gli errori da remoto
				// ...
			} catch (OpCodeNotPresentException e) {
				e.printStackTrace();
				result = null; // TODO cosi' non si vedono gli errori da remoto
				// ...
			} catch (Exception e) {
				log.warn("Exception " + e + " while computing mdfi");
				e.printStackTrace();
			}

			return result;
		}
	}

	/**
	 * @param args
	 */

	public static void main(String[] args) throws RemoteException {
		int port;
		if (args.length != 0)
			port = Integer.parseInt(args[0]);
		else
			port = Port;
		System.out.println("==========================================\n");
		try {
			System.out
					.println("Working on host: " + InetAddress.getLocalHost());
		} catch (UnknownHostException e2) {
			e2.printStackTrace();
		}
		System.out.println("JRE Version:          "
				+ System.getProperty("java.version"));
		System.out.println("OS Information:       "
				+ System.getProperty("os.name") + " "
				+ System.getProperty("os.version") + " "
				+ System.getProperty("os.arch"));
		System.out.println("User Login:           "
				+ System.getProperty("user.name"));
		System.out.println("==========================================\n");
		if (System.getProperty("java.version").compareTo("1.6") < 0) {
			System.err.println("!!!WARNING: Use JDK version 1.6 or higher!!!");
			System.exit(-1);
		}
		if (args.length != 0) {
			if (args[0].equals("-security")) {
				RMISecurityManager rmiSM = new RMISecurityManager();
				System.setSecurityManager(rmiSM);
			}
		}

		RemoteInterpreterInterface worker = new RemoteInterpreter();
		log.info("Worker created ..");
		Registry reg = null;
		try {
			reg = LocateRegistry.createRegistry(port); // or I'm the first one
			// running on this
			// server
		} catch (RemoteException e) {
			e.printStackTrace();
			System.exit(-1);//port already in use, exit.
		}

		log.info("Register created on port " + port);
		// now register the object
		try {
			String workerName = "muSkelMDFWorker";
			reg.rebind(workerName, (Remote) worker);
		} catch (AccessException e1) {
			e1.printStackTrace();
		} catch (RemoteException e1) {
			e1.printStackTrace();
		}
		log.info("muskel Worker registered ...");
		// now start the presence thread ...
		PresenceThread pt = new PresenceThread();
		pt.start();
		log.info("Presence thread started ...");
	}

}