/* This is the muskel: a skeleton/RMI based, parallel programming library
Copyright (C) 2006  Marco Danelutto, Dept. Computer Science, Univ. Pisa, Italy

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA. */

package muskel.mdf;

import java.io.Serializable;

/**
 * used to store a destination in compact form.
 * A triple InputPositionId, InstructionId, GraphId
 * 
 * @author marcod
 * @author albanese
 */
public class Destination implements Serializable, Cloneable {

	private static final long serialVersionUID = -498727500300417353L;
	
	private InputPositionId p = new InputPositionId(new IntegerIdentifier(0)); // position in the input token vector
	private InstructionId i = Mdfi.NoInstrId;
	private GraphId g = Mdfi.NoGraphId;
	
	/**
	 * The Destination constructor
	 * 
	 * @param pp position of the dest token in the input vector
	 * 		of the target instruction (positions start at 0)
	 * @param ii identifier of the target instruction.
	 * 			In case the token is destined to the output stream,
	 * 			this must be a Mdfi.NoInstrId constant.
	 * @param gg graph identifier of the target instruction.
	 * 			While constructing the graph, this should be set to Mdfi.NoGraphId
	 */
	public Destination(InputPositionId pp, InstructionId ii, GraphId gg) {
		p = pp;
		i = ii; 
		g = gg;
	}
	/**
	 * The Destination constructor, to be used when constructing MDF graphs.
	 * In this case there is no need to specify the graph id, 
	 * as it will be "relocated" later
	 * 
	 * @param pp position of the Destination token in the input vector
	 * 			of the target instruction (positions start at 0 ???)
	 * @param ii identifier of the target instruction. 
	 * 		In case the token is destined to the output stream,
	 * 		this is to be a Mdfi.NoInstrId constant.
	 */
	public Destination(InputPositionId pp, InstructionId ii) {
		this(pp, ii, Mdfi.NoGraphId);
	}
	
	/**
	 * Returns the position in the destination instruction
	 * input token vector
	 * @return the position in the destination instruction
	 * 			input token vector
	 */
	public InputPositionId getInputPositionId() {
		return p;
	}
	/**
	 * Accessory method
	 * @return the destination instruction id
	 */
	public InstructionId getInstructionId() {
		return i;
	}
	
	 /**
	  * accessory method
	  * @return the graph id of the target instruction
	  */
	public GraphId getGraphId() {
		return g;
	}
	
	public void setInputPositionId(InputPositionId p) {
		this.p = p;
	}
	public void setInstructionId(InstructionId i) {
		this.i = i;
	}
	public void setGraphId(GraphId g) {
		this.g = g;
	}
	
	/**
	 * overwrites Object toString method
	 * @return the string representation of the Destination object
	 */
	public String toString() {
		return ("Destination : <"+p+","+i+","+g+">");
	}
	
	/**
	 * Deep copy
	 * */
	@Override
	public Object clone() {

		InputPositionId cloneP = (InputPositionId) this.p.clone();
		InstructionId cloneI = (InstructionId) this.i.clone();
		GraphId cloneG = (GraphId) this.g.clone();
		
		return new Destination(cloneP, cloneI, cloneG);
	}

}
