/* This is the muskel: a skeleton/RMI based, parallel programming library
Copyright (C) 2006  Marco Danelutto, Dept. Computer Science, Univ. Pisa, Italy

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA. */

package muskel.mdf;

import java.util.Collection;
import java.util.HashMap;
import java.util.Map;
import muskel.Manager;

/**
 * This is the class used to store a whole graph.
 * Accessory methods are provided to interact with and use the graph.
 * 
 * The Input Mdfi must have an unique input token.
 * 
 * The Output Mdfi must have an unique Destination and
 * the Destination InstructionId must be Mdfi.NoInstrId
 * 
 * Usage example:
 * 
 * 		MdfGraph graph = new MdfGraph();
 * 
 * 		graph.setInputInstruction(i1);
 *		graph.addInstruction(i1);
 *		graph.addInstruction(i2);
 *		graph.addInstruction(i3);
 *		graph.addInstruction(i4);
 *		graph.setOutputInstruction(i4);
 * 
 * @author marcod
 * @author albanese
 */
public class MdfGraph {

	public final boolean debug = Manager.debug;
	
	Map<InstructionId, Mdfi> graph = null;
	// used to store the instructions
	
	/**
	 * L'id della prima mdfi del grafo
	 * deve avere un unico token di input
	 * */
	InstructionId inputMdfiId = null;
	
	/**
	 * L'id dell'ultima mdfi del grafo l'unica
	 * che ha NoIstructionId come id di destinazione
	 * deve avere, inoltre, un unico token di uscita
	 * */
	InstructionId outputMdfiId = null;
	
	GraphId gid = Mdfi.NoGraphId;  // this is the graph id, actually
	
	/**
	 * numero di mdfi con SECompute
	 * */
	private int numSECompute = 0;
		
	/**
	 * Standard constructor
	 */
	public MdfGraph() {
		this(Mdfi.NoGraphId);
	}

	/** 
	 * Constructor with a given graph id (not used at the moment)
	 * @param ggid
	 */
	public MdfGraph(GraphId ggid) {
		graph = new HashMap<InstructionId, Mdfi>();
		gid = ggid;
	}
	
	
	/**
	 * Getter method for the graph
	 */
	public Map<InstructionId, Mdfi> getGraph() {
		return graph;
	}
	
	/**
	 * Returns a collection containing the graph Mdfi
	 */
	public Collection<Mdfi> getGraphMdfi() {
		return graph.values();
	}

	/**
	 * used to set the graph id when the graph is instantiated
	 * @param ggid
	 */
	public void setId(GraphId ggid) {
		
		gid = ggid;
		
		/*ottengo una vista dei valori del Map come
		 * Collection in modo da poter iterare
		*/
		Collection<Mdfi> iterableGraph = graph.values();
		
		//aggiorno nelle mdfi il graphId
		for (Mdfi mdfi : iterableGraph) {
			mdfi.setGraphId(ggid);
		}
		
	}

	/**
	 * this is used to get the graph id.
	 * It will be used to instantiate new instructions ...
	 * not derived from skeleton compilation
	 * @return the id of the graph
	 */
	public GraphId getGraphId() {
		return gid;
	}

	/**
	 * Return the graph's SE Mdfi number
	 * 
	 * */
	public int getNumSECompute() {
		return numSECompute;
	}

	/**
	 * Adds an instruction to the graph
	 * @param instr the instruction to be added
	 */
	public void addInstruction(Mdfi instr) {
		
		//aggiungo usando come chiave l'instructionId di instr
		graph.put(instr.getInstrId(), instr);
		
		//se instr e' ha side effects aggiorno il relativo contatore
		if(instr.isSeMDFi())
			numSECompute++;
	}
	
	/**
	 * Sets the first graph instruction.
	 * A Mdfi is a valid Input Mdfi if:
	 * - it has an unique InputToken
	 * 
	 * @param intputInstr The instruction to be set
	 * @throws IllegalArgumentException If inputInstr
	 * 			is null or if it doesn't have an unique input Token
	 */
	public void setInputInstruction(Mdfi intputInstr) {
		
		if(!validateInputMdfi(intputInstr))
			throw new IllegalArgumentException();
		
		//aggiungo usando come chiave l'instructionId di firstInstr
		inputMdfiId = intputInstr.getInstrId();
		
	}
	
	/**
	 * Sets the last instruction to the graph.
	 * A Mdfi is a valid Output Mdfi if:
	 * - it has an unique Destination
	 * - the Destination Instruction Id must be Mdfi.NoInstrId
	 * 
	 * @param outputInstr the instruction to set
	 * @throws IllegalArgumentException If outputInstr is null or if it isn't valid.
	 */
	public void setOutputInstruction(Mdfi outputInstr) {
		
		if(!validateOutputMdfi(outputInstr))
			throw new IllegalArgumentException();
		
		//aggiungo usando come chiave l'instructionId di outputInstr
		outputMdfiId = outputInstr.getInstrId();
		
	}
	
	/**
	 * Controlla se una mdfi puo' essere usata come mdfi di input di un grafo.
	 * Essa deve avere un unico Token di Input
	 * 
	 * */
	private boolean validateInputMdfi(Mdfi intputInstr) {
				
		if (intputInstr == null)
			return false;

		//La Mdfi di input deve avere un unico input Token.
		if (intputInstr.getInCount() != 1)
			return false;
		
		return true;
	}
	
	/**
	 * Controlla se una mdfi puo' essere usata come mdfi di output di un grafo
	 * 
	 * A Mdfi is a valid Output Mdfi if:
	 * - it has an unique Destination
	 * - the Destination Instruction Id is Mdfi.NoInstrId
	 * 
	 * */
	private boolean validateOutputMdfi(Mdfi outputInstr) {
		
		if (outputInstr == null)
			return false;
		
		//La Mdfi di output deve avere un unico output Token.
		if (outputInstr.getOutCount() != 1)
			return false;
		
		//prendo il primo (e unico, come appena controllato) outputToken della mdfi
		OutputToken outToken = outputInstr.getOutTokenVector()[0];
		
		//prelevo la testinazione dal token
		Destination dest = outToken.getDestination();
		
		//l'InstructionId contenuto in dest DEVE essere NoInstrId
		if(! (dest.getInstructionId().equals(Mdfi.NoInstrId)) )
			return false;
		
		//tutto ok
		return true;
	}
	
	/**
	 * This is used to get the id of the input instruction,
	 * the one where the input token has to be stored
	 * @return the id of the first instruction
	 */
	public InstructionId getInputInstrId() {
		return inputMdfiId;
	}
	
	/**
	 * This is used to get the id of the output instruction
	 * @return the id of the last instruction
	 */
	public InstructionId getOutputInstrId() {
		return outputMdfiId;
	}

	/**
	 * Stores he given destination, if not null, in the last instruction
	 * (i.e. the only one whose unique output token dest is NoInstrId).
	 * @param dest the Destination to be given to the last instruction
	 * @return the InstructionId of the last instruction
	 * @throws IllegalArgumentException if dest is null
	 */
	public InstructionId setOutputInstrId(Destination dest) {
		
		if(dest == null)
			throw new IllegalArgumentException();
		
		if(debug) System.out.println("setOutput Dest is "+dest.toString());
		if(debug) System.out.println("IN GRAPH "+graph.toString());
		
		Mdfi outputMdfi = graph.get(outputMdfiId);
		
		if(debug) System.out.println("CHANGING "+outputMdfi);
			
		int firstPosition = 0;//prima posizione nell'array delle destinazioni
		
		/*aggiorno la prima e unica 
		 * essendo, essa, una mdfi di output) destinazione.
		*/outputMdfi.setDestination(firstPosition, dest);
		
		if(debug) System.out.println("BECOMES "+outputMdfi);
			
		return(outputMdfiId);
	}

	/**
	 * Used to set graph id to a given value.
	 * It is used when a graph is instantiated with the new input stream token
	 * @param ggid the graph id
	 * @throws IllegalArgumentException If ggid is null
	 *  */
	public void setGid(GraphId ggid) {
		
		if(ggid == null)
			throw new IllegalArgumentException();
		
		/*ottengo una vista dei valori del Map
		 * come Collection in modo da poter iterare
		*/Collection<Mdfi> iterableGraph = graph.values();
		
		for (Mdfi mdfi : iterableGraph) {
			mdfi.setGraphId(ggid);
			mdfi.setDestGraphIds(ggid);
		}
		
		gid = ggid;

	}

	/** 
	 * This is the method computing an MDF Graph on a given input token. <br>
	 * @param task the value of the input task
	 * @return the value of the unique computed token
	 */

	public Task[] compute(Task[] task) {
		// TODO : this is dummy !
		Task[] dummy = null;
		System.out.println("Unimplemented compute method in MdfGraph");
		return dummy;
	}
	
	/**
	 * overrides Object toString
	 * @return the pretty printed graph
	 */
	public String toString() {
		String out = "GRAPH is\n";
		
		/*ottengo una vista dei valori del Map
		 *come Collection in modo da poter iterare
		*/
		Collection<Mdfi> iterableGraph = graph.values();
		
		int i = 0;
		
		for (Mdfi mdfi : iterableGraph) {
			out = out + "instr:"+i+":"+mdfi+"\n";
			i++;
		}
			
		return out;
	}

	/**
	 * Controlla se il grafo è stato inizializzato.
	 * Cioè se sono state definite le mdfi di in e/o di out
	 * 
	 * */
	private boolean isInitialized()
	{
		return (inputMdfiId != null) && (outputMdfiId != null);		
	}
	
	
	/**
	 * overrides the Object clone.
	 * Used to create a new instance of the graph
	 * each time a new item is got from the input stream.
	 * @return the (empty) graph
	 * @throws IllegalStateException Se il grafo non e' stato
	 * 				inizializzato correttamente
	 * 					(non sono definite le mdfi di in e/o di out)
	 */
	public MdfGraph mdfClone() {
		
		//controllo se sono state definite le mdfi di in e di out
		if(!isInitialized())
			throw new IllegalStateException("Graph not initialized.");
			
		//creo un grafo vuoto
		MdfGraph ng = new MdfGraph();
		
		ng.graph = new HashMap<InstructionId, Mdfi>();
			
		/*ottengo una vista dei valori del Map come Collection
		 * in modo da poter iterare
		*/
		Collection<Mdfi> iterableGraph = this.graph.values();
		
		//mdfi di input
		Mdfi toCloneInputInstr = this.graph.get(this.getInputInstrId());
		
		//mdfi di output
		Mdfi toCloneOutputInstr = this.graph.get(this.getOutputInstrId());
				
		//clono le mdfi e le aggiungo al nuovo grafo	

		//clono l'istruzione di input
		Mdfi clonedInputInstr = toCloneInputInstr.mdfClone();
		//la aggiungo al grafo
		ng.addInstruction(clonedInputInstr);
		//la imposto come istruzione di input del grafo clonato
		ng.setInputInstruction(clonedInputInstr);
				
		//poi le altre tranne quella di output
		for (Mdfi mdfi : iterableGraph) {
			
			// non è una mdfi ne' di in ne' di out
			if(toCloneInputInstr != mdfi && toCloneOutputInstr != mdfi) {
				//clono
				Mdfi clonedMdfi = mdfi.mdfClone();
				//aggiungo
				ng.addInstruction(clonedMdfi);
			}
		}
		
		//clono l'istruzione di output
		Mdfi clonedOutputInstr = toCloneOutputInstr.mdfClone();
		//la aggiungo al grafo
		ng.addInstruction(clonedOutputInstr);
		//la imposto come istruzione di output del grafo clonato
		ng.setOutputInstruction(clonedOutputInstr);
		
		//clono il graphId
		ng.gid = (GraphId) this.gid.clone();
		
		return ng;
	}
	

	public Object clone() {
		
		return mdfClone();
	}
}
