/* This is the muskel: a skeleton/RMI based, parallel programming library
Copyright (C) 2006  Marco Danelutto, Dept. Computer Science, Univ. Pisa, Italy

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA. */

package muskel.mdf;

import java.io.Serializable;
import skeleton.Compute;
import skeleton.SECompute;
import muskel.CodeStorage;
import muskel.Manager;
import muskel.NonFireableInstructionException;
import muskel.OpCodeNotPresentException;
import muskel.TokenNotPresentException;

/**
 * This is the class of MDF instructions: each MDFi has a
 * instrId opCode tokenInVector tokenOutVector readyArgNo inTokenCount outTokenCount
 * @author marcod
 * @author albanese
 *
 */
public class Mdfi implements Serializable {
	
	public final boolean debug = Manager.debug;

	private static final long serialVersionUID = 1L;
	
	private static final Identifier NO_ID = IntegerIdGenerator.getNoId();
	
	public final static InstructionId NoInstrId = new InstructionId(NO_ID);
	public final static GraphId NoGraphId = new GraphId(NO_ID);
	public final static OperationId NoOpCode = new OperationId(NO_ID);
	
	private InstructionId instrId = NoInstrId;
	private GraphId graphId = NoGraphId;
	private OperationId opCode = NoOpCode;
	
	InputToken [] inTokenVector = null;

	Destination [] dests = null;
	
	OutputToken [] outTokenVector = null;
	
	int readyArgNo = 0;
	int inCount = 0;
	int outCount = 0;
	
	boolean fireable = false;
	
	/**flag per identificare le mdfi con side effects*/
	boolean seMDFi = false;
		
    /** Creates a MDFi taking care of storing the code in the CodeStorage 
     * This is used only in case of user defined ParCompute code. 
     * Same parameters of the previous, complete constructor,
     * but the GID, which is left undefined. 
     * Plus the manager, which is needed to access the global CodeStorage
     * @param manager the manager used to compute the instruction
     * @param iid the instruction id
     * @param op the code to be computed by the instruction. 
     * 			Should be a Compute (sub)class object
     * @param inc the input token number
     * @param outc the output token number
     * @param did the vector of the destinations for output tokens.
     * 			The i-th output token is directed to the i-th Destination
     */ 
	public Mdfi(Manager manager, InstructionId iid, Compute op, int inc, int outc, Destination[] did) {

		// first set up all parameters but the opcode checking the op type
		this(iid,NoGraphId, NoOpCode, inc, outc, did, (op instanceof SECompute));
		// then fix the opcode, after storing the Compute code in the CodeStore
		opCode = manager.storeOpCode(op);
		
		if(debug) System.out.println("first constructor: outToken dims to "+outTokenVector.length);
	}

   /** 
    *
    * Creates a MDFi
    * @param iid the instruction id
    * @param op The code id in the CodeStorage to be
    * 			computed by the instruction. 
    * @param inc the input token number
    * @param outc the output token number
    * @param did the vector of the destinations for output tokens.
    * 			 The i-th output token is directed to the i-th Destination
    * @param seMDFi the flag representing a mdfi having a side action compute
    */ 
	public Mdfi(InstructionId iid, GraphId gid,
			OperationId op, int inc, int outc, Destination[] did, boolean seMDFi) {
		
		this.instrId = iid;
		this.graphId = gid;
		this.opCode = op; 
		inCount = inc;
		outCount = outc;
		this.dests = did;
		
		this.seMDFi = seMDFi;
		
		inTokenVector = new InputToken[inc];
		for(int i=0; i<inCount; i++)
			inTokenVector[i] = new InputToken();
			// initialize the Token input vector (all empty at the moment)

		outTokenVector = new OutputToken[outCount];
		for(int i=0; i<outCount; i++)
			outTokenVector[i] = new OutputToken(dests[i]);
		
		readyArgNo = 0;
		fireable = false;

		if(debug)
			System.out.println("first constructor: outToken dims to "
												+outTokenVector.length);
	}
	
	/**
	 * Used to store an input token
	 * @param position the position in the input token vector
	 * @param value the value of the token
	 */
	public void storeToken(InputPositionId position, Task value) {
		//ricavo l'int dall InputPositionId
		int pos = ((IntegerIdentifier) position.getId()).getId();
		
		inTokenVector[pos].updateToken(value);
		readyArgNo++;
		if(readyArgNo==inCount)
			fireable = true;
	}
	
	/** 
	 * Computes the MDF instruction, taking the opcode from the CodeStorage
	 * @param cs the CodeStorage to be used to retrieve instruction code
	 * @return the vector of the result tokens (OutputTokens, actually)
	 * @throws NonFireableInstructionException
	 * @throws OpCodeNotPresentException
	 */
	public ComputeResult compute(CodeStorage cs) throws NonFireableInstructionException, OpCodeNotPresentException  {
		// log.debug("MDFI COMPILE: going to compute");
		if(fireable) {
			// log.debug("MDFI COMPILE: going to compute fireable");
			// first compute the values
			Task[] tokenValues = new Task[inCount];
			for(int i=0; i<inCount; i++) {
				try {
					tokenValues[i] = inTokenVector[i].getTask();
				}catch(TokenNotPresentException e) {
					throw new NonFireableInstructionException();
				} catch(ArrayIndexOutOfBoundsException e) {
				    e.printStackTrace();
				}
			}
			// log.debug("MDFI COMPILE: computing fireable: got tokens");
			if(debug) System.out.println("going to compute with "+tokenValues.length+" input tokens ");
			
			Compute comp = cs.getCompute(opCode);
			
			Task[] result;
			ComputeResult toReturn;
			
			Task[] generatedTask = null;
			if(seMDFi) {
				/*
				 * nel caso di un SEcompute e' necessario
				 * rendere atomicala sequenza delle
				 * seguenti 2 operazioni:
				 * - compute(), dove viene impostata la
				 * 	 variabile condivisa generatedTask,
				 * - getGeneratedTask() dove viene prelevata
				 * 	 la variabile stessa.
				 * 
				 * Devo, percio', sincronizzarmi su comp
				 * (contenuto nel CodeStorage)
				 * che contiene la variabile condivisa.
				 * Cosi' facendo sono sicuro di essere
				 * l'unico ad operare su comp.
				*/
				synchronized (comp){
					result = (Task[]) comp.compute(tokenValues);
					
					SECompute seComp = (SECompute) comp;
					
					Task[] gtTemp = seComp.getGeneratedTask();
					
					if(gtTemp != null) {
												
						generatedTask = new Task[gtTemp.length];
						//clono i task generati
						for(int i = 0; i< gtTemp.length; i++)
							generatedTask[i] = (Task) gtTemp[i].clone();
						
						//reimposto il valore per il prossimo utilizzo
						seComp.resetGeneratedTask();
					}
				}
				
			}
			else {//caso Compute
				result = (Task[]) comp.compute(tokenValues);
			}			
			/*  ATTENZIONE  L'i-esimo valore viene passato all'i-esima destinazione.
			 *  Quindi una mdfi con n destinazioni DEVE avere
			 *  una Compute che calcoli con n output.
			 *  E' a carico di chi scrive la Mdfi controllare che il num
			 *  di destinazioni sia compatibile (cioè #dest = #result)
			 *  con la Compute utilizzata.
			 *  */

			//i-esimo risultato all'i-esima destinazione
			for(int i=0; i<outCount; i++) 
				outTokenVector[i] = new OutputToken(result[i], dests[i]);			
			
			//genero il risultato in base al tipo di mdfi
			if(seMDFi)
				toReturn = new SEComputeResult(outTokenVector, generatedTask);			
			else toReturn = new ComputeResult(outTokenVector);
			
			// eventually return the OutputToken vector
			return toReturn;
			
		} else 
			throw new NonFireableInstructionException();
	}
	
	/**
	 * test for fireable instruction
	 * @return returns true if all the input
	 * tokens are present and the instruction can be computed
	 */
	public boolean isFireable() {
		return fireable;
	}	

	public InstructionId getInstrId() {
		return instrId;
	}


	public GraphId getGraphId() {
		return graphId;
	}


	public InputToken[] getInTokenVector() {
		return inTokenVector;
	}
	
	public OutputToken[] getOutTokenVector() {
		return outTokenVector;
	}
	
	
	/**
	 *Il un numero di token di ingresso della Mdfi
	 */
	public int getInCount() {
		return inCount;
	}

	/**
	 *Il un numero di token di uscita della Mdfi
	 */
	public int getOutCount() {
		return outCount;
	}
	
	public boolean isSeMDFi() {
		
		return seMDFi;
	}
	
	/**
	 * Assigns a graph id to the instruction, possibly created with a Mdfi.NoGraphId value. 
	 * @param gid the graph id to be assigned
	 * @throws IllegalArgumentException If gid is null
	 */
	public void setGraphId(GraphId gid) {
		
		if (gid == null)
			throw new IllegalArgumentException();
		
		graphId = gid;
	}	
	
	/**
	 * Sets the graph id in the destination tokens
	 * @param gid the graph id to be used
	 */
	public void setDestGraphIds(GraphId gid) {
		for(int i=0; i<outCount; i++) {
			 //cosi' modifico anche outputToken perchè punta all'i-esimo dest
			dests[i].setGraphId(gid);
			
			//dests[i] = new Destination(dests[i].getInputPositionId(), dests[i].getInstructionId(), gid);
			//outTokenVector[i].setDest(dests[i]);
		}
	}
	
	/**
	 * Modifica la destinazione del ppos-esimo token di output
	 * @param ppos the index of the destination to be changed
	 * @param dest the new destination
	 */
	public void setDestination(int ppos, Destination dest) {
		
		this.dests[ppos] = dest;
		this.outTokenVector[ppos] = new OutputToken(dest);

		if(debug) System.out.println("Changing dest pos "+ppos+" with "+dest.toString()+"\ngives: "+this.toString());

	}
	
	/**
	 * Overrides the Object toString
	 * @return the pretty printed instruction
	 */
	public String toString() {
		String out = "MDFI <"+instrId+","+graphId+">:";
		out = out + "op="+opCode+":in="+inCount+":out="+outCount+":fireable="+fireable+":tokens:";
		for(int i=0; i<inCount; i++)
			out = out + inTokenVector[i].toString();
		out = out + ":destTokens:";
		for(int i=0; i<outCount; i++)
			out = out + outTokenVector[i].toString();
		return out;
	}
	
	/**
	 * used by clone()
	 * */
	private Mdfi() {}
	
	/**
	 * This is used to clone an instruction. 
	 * Called by mdfClone in MdfGraph, when a new instance
	 * of the graph is to be stored with the input
	 * token from the input stream as the graph input token. 
	 * <br>
	 * Does not clone the opcode, as this is data flow
	 * 
	 * @return a copy of the current instruction
	 */
	public Mdfi mdfClone() {
		
		Mdfi ni = new Mdfi();
		
		ni.instrId = (InstructionId) instrId.clone();
		ni.graphId = (GraphId) graphId.clone();
		ni.opCode = this.opCode;
		
		ni.inTokenVector = (inTokenVector == null) ? null : cloneInTokenArray();		
		ni.dests = (dests == null) ? null : cloneDestinationArray();		
		ni.outTokenVector = (outTokenVector == null) ? null : cloneOutputTokenArray();
		
		ni.readyArgNo = readyArgNo;
		ni.inCount = inCount;
		ni.outCount = outCount;
		ni.fireable = fireable;
		ni.seMDFi = seMDFi;
		
		return ni;
	}
	
	/**
	 * Clona con deep copy il vettore dei token d'ingresso
	 * */
	private InputToken[] cloneInTokenArray()
	{
		InputToken[] clone = new InputToken[this.inTokenVector.length];
		for(int i=0; i<this.inTokenVector.length;i++) 
			clone[i] = (InputToken) this.inTokenVector[i].clone();
		
		return clone;
	}

	/**
	 * Clona con deep copy il vettore dei token d'uscita
	 * */
	private OutputToken[] cloneOutputTokenArray()
	{
		OutputToken[] clone = new OutputToken[this.outTokenVector.length];
		for(int i=0; i<this.outTokenVector.length;i++) 
			clone[i] = (OutputToken) this.outTokenVector[i].clone();
		
		return clone;
	}
	
	/**
	 * Clona con deep copy il vettore delle destinazioni
	 * */
	private Destination[] cloneDestinationArray()
	{
		Destination[] clone = new Destination[this.dests.length];
		for(int i=0; i<this.dests.length;i++) 
			clone[i] = (Destination) this.dests[i].clone();
		
		return clone;
	}
	
}