package muskel.stream;

import java.util.concurrent.LinkedBlockingQueue;
import org.apache.log4j.Logger;

/**
 * Wrapper per un NonBlockingInputManager che sono "statici",
 * cioe' che non permettono l'aggiunta di elementi da produrre a runtime.
 * 
 * DynamicInputManager consuma, durante la sua creazione,
 * NonBlockingInputManager (che dovra' contenere almeno un elemento)
 * passato per parametro; inoltre permette a dei produttori
 * in inserire nello stream nuovi elementi.
 * 
 * I produttori devono segnalare allo stream sia
 * la volonta' di iniziare a produrre, che quella di smettere.
 * 
 * */
public class DynamicInputManager implements BlockingInputManager {
	
	private LinkedBlockingQueue<Object> queue;
	
	/*numero di produttori dichiaratisi interessati a produrre*/
	private int nProducers;
	
	private boolean closed;
		
	private static Logger log = Logger.getLogger(DynamicInputManager.class.getName());
	

	/**
	 * Crea un nuovo stream usando riempiendolo con gli elementi
	 * di inStream, esso deve conterne almeno uno.
	 * @param inStream uno stream contentente almeno un elemento
	 * */
	public DynamicInputManager(NonBlockingInputManager inStream)	{
		
		queue = new LinkedBlockingQueue<Object>();
		
		closed = false;
		//uso -1 per evitare l'ambiguita' dello 0: stream chiuso o mai usato?
		//alla prima addProducers nProducers conterra' il numeto effettivo di produttori
		nProducers = -1;
	
		//inserisco nella coda gli elementi dello stream
		Object task; 
		while ((task = inStream.next()) != null) {
			queue.add(task);
			log.debug("Accodato task ottenuto dallo stream iniziale: "+ task);
		}
	}

	/**
	 * Restituisce true se ci sono elementi pronti da consumare,
	 * si blocca se non ci sono elementi pronti ma sono presenti
	 * potenziali produttori. Qualora i produttori non aggiungano
	 * nessun elemento durante l'attesa,il metodo restituisce false
	 * */
	@Override
	public synchronized boolean hasNext() {
		
		if(!closed && nProducers > 0) {
			/*
			 * lo stream non e' stato ancora chiuso, cioe' 
			 * non ho mai avuto zero consumatori
			 * 
			 * Se lo stream iniziale contiene almeno un elemento,
			 * e this viene usato da compute produttori
			 * (SECompute), esso avra' 0 produttori solo al termine dell'esecuzione  
			 * */		
			
			if (!queue.isEmpty()) {
				log.trace("La coda NON e' vuota ci sono task da consumare");
				return true; //la coda NON e' vuota ci sono task da consumare
			}
			else {
				//la coda e' vuota, ma potrebbe riempirsi, aspetta
				while(queue.isEmpty()) {
					
					if(closed) {
						log.trace("I produttori che attendevamo non hanno prodotto, lo stream e' chiuso");
						/*Durante la wait i produttori
						 * che attendevamo non hanno prodotto,
						 * lo stream e' chiuso
						*/return false;
					}
					
					try {
						log.trace("La coda e' vuota, ma potrebbe riempirsi, aspetta...");
						
						wait();			
						
						log.trace("...Svegliato");
					} catch (InterruptedException e) {e.printStackTrace();}
					
				}
				
				log.trace("La coda e' stata riempita. Posso estrarre.");
				return true;//la coda e' stata riempita ci sono task da consumare
			}
		}
		else {//non ci sono produttori (stream chiuso) ma la coda potrebbe non essere vuota

			log.trace("Non ci sono produttori, potrebbero esserci task rimanenti");

			if (queue.isEmpty()) {
				log.trace("Non ci sono altri task, lo stream e' esaurito");
				return false; //la coda e' vuota, lo stream e' esaurito
			}
			else {
				log.trace("Ci sono task da consumare");
				return true; //la coda NON e' vuota, ci sono task da consumare
			}
		}
	}
		

	@Override
	public Object next() {
		
		//se la coda e' vuota aspetta
		Object toReturn = null; 
		try {
			toReturn = queue.take();
			log.debug("Task ottenuto: " + toReturn);	
		} catch (InterruptedException e) {
			e.printStackTrace();
		}
		
		return toReturn;
	}
	
	public synchronized void addTask(Object task) {
		
		try {
			//se c'era qualcuno in attesa di un task lo sveglio
			queue.put(task);
			log.debug("Accodato task, totale: "+ queue.size());	
		} catch (InterruptedException e) {}//eccezione se non c'e' spazio della coda, improbabile
		notifyAll();//sveglio i consumatori che si sono messi in attesa durante la hasNext()
	}
	
	public synchronized void addProducers(int n) {
		
		if(!closed) {
			
			if(nProducers == -1) {//e' il primo produttore che aggiungo
				nProducers = n;
				log.trace("Aggiunto primo produttore, totale: "+ nProducers);
			}
			else {
				nProducers += n;
				
				if(n==1)
					log.trace("Aggiunto un produttore, totale: "+ nProducers);
				else log.trace("Aggiunti " + n +" produttori, totale: "+ nProducers);
			}
		}
		else throw new IllegalStateException("Stream Closed");
	}
	
	public synchronized void addProducer() {
		
		if(!closed) {
			
			if(nProducers == -1) {//e' il primo produttore che aggiungo
				nProducers = 1;
				log.trace("Aggiunto primo produttore, totale: "+ nProducers);
			}
			else {
				nProducers++;
				log.trace("Aggiunto produttore, totale: "+ nProducers);		
			}
		}
		else throw new IllegalStateException("Stream Closed");
	}
		
	
	public synchronized void removeProducer() {		
		
		if(!closed) {
			
			//decremento il num dei produttori se questo e' maggiore di 0
			if (nProducers > 0) {
				nProducers--;
				log.trace("Rimosso produttore, totale: "+ nProducers);
				
				if (nProducers == 0) {	
					/*il numero di produttori e' arrivato a zero
					 * quindi ho finito e dichiaro lo stream chiuso
					*/
					closed = true;
					notifyAll();
					/*sveglio i consumatori in attesa di nuovi
					 * probabili task durante la hasNext()
					 * ma siccome sono tutti terminati non
					 * arriveranno nuovi task
					 */
					log.trace("Stream chiuso");
				}
			}
			else //nProducers <= 0
				throw new IllegalStateException("Non ci sono produttori da eliminare");
			
		}
		else throw new IllegalStateException("Stream Closed");
	}
}