/* This is the muskel: a skeleton/RMI based, parallel programming library
 Copyright (C) 2006  Marco Danelutto, Dept. Computer Science, Univ. Pisa, Italy

 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 as published by the Free Software Foundation; either version 2
 of the License, or (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA. */


package skeleton;

import muskel.mdf.MdfGraph;
import muskel.mdf.Task;

/**
 * This is the class used to provide user defined MDF graphs as (parts of)
 * skeletons. <br>
 * 
 * ParCompute items can be used where Compute items are. Unlike Compute items,
 * requiring the user provides the Task[] compute(Task[]) method, the ParCompute
 * items must be created providing a MdfGraph macro data flow graph, with a
 * single token in input and a single token in output. <br>
 * The computation of the ParCompute goes through the supplied graph rather than
 * compiling MDF instructions using the compute method as opcode.<br>
 * 
 *<p>
 *Sample use is as follows:
 *
 *<pre>
		Manager manager = new Manager(); 

		Compute inc1 = new Splitter();
		
		IntegerIdGenerator intIdGen = new IntegerIdGenerator();
		
		InstructionId i1_Id = new InstructionId(intIdGen.generateId());
		InstructionId i2_Id = new InstructionId(intIdGen.generateId());
		InstructionId i3_Id = new InstructionId(intIdGen.generateId());
		InstructionId i4_Id = new InstructionId(intIdGen.generateId());
		
		Destination d1 = new Destination(
				new InputPositionId(new IntegerIdentifier(0)), i2_Id, Mdfi.NoGraphId);

		Destination d2 = new Destination(
				new InputPositionId(new IntegerIdentifier(0)), i3_Id, Mdfi.NoGraphId);
				
		Destination[] dests = {d1, d2};
		Mdfi i1 = new Mdfi(manager,i1_Id,inc1,1,2,dests);
		
		Compute sq1 = new Square();
		
		Destination d3 = new Destination(
				new InputPositionId(new IntegerIdentifier(0)), i4_Id, Mdfi.NoGraphId);
		
		Destination[] dests1 = {d3};
		
		Mdfi i2 = new Mdfi(manager, i2_Id, sq1, 1, 1, dests1);
		
		Compute sq2 = new Square();
		
		Destination d4 = new Destination(
				new InputPositionId(new IntegerIdentifier(1)), i4_Id, Mdfi.NoGraphId);
		
		Destination[] dests2 = {d4};
		Mdfi i3 = new Mdfi(manager, i3_Id, sq2, 1, 1, dests2);
		
		Compute add = new Comp();
		
		Destination d5 = new Destination(
				new InputPositionId(new IntegerIdentifier(0)), Mdfi.NoInstrId, Mdfi.NoGraphId);
		
		Destination[] dests3 = {d5};
		
		Mdfi i4 = new Mdfi(manager, i4_Id, add, 2, 1, dests3);		
				
		MdfGraph graph = new MdfGraph();
		
		graph.setInputInstruction(i1);
		graph.addInstruction(i1);
		graph.addInstruction(i2);
		graph.addInstruction(i3);
		graph.addInstruction(i4);
		graph.setOutputInstruction(i4);
		
		ParCompute main = new ParCompute(graph);

 </pre>
 * In this case a graph not expressible with the standard muskel
 * skeletons is built. <br>The graph can be used in a farm,
 * as an example, by telling
 <pre>
 Farm f = new Farm(mdfg);
 manager.setProgram(f);
 </pre>
 * or computed as is by telling
 <pre>
 manager.setProgram(mdfg);
 </pre>
 * just before issuing the <code> manager.compute()</code> call.
 *
 * @author marcod
 */
public class ParCompute extends Compute {

	private static final long serialVersionUID = 1L;
	
	protected MdfGraph program = null;

	/**
	 * ParCompute nodes are created by passing an MDF graph as the node
	 * "program"
	 * 
	 * @param g
	 *            the graph to be computed
	 */
	public ParCompute(MdfGraph g) {
		program = g;
	}

	/**
	 * This is the method that has to be computed in order to implement the
	 * Compute interface. It is currently null (actually, it always returns a
	 * null, but it can be derived interpreting on the fly the MDF graph.<br>
	 * Users can extend the class providing an actual method, but this will
	 * not be used but for computing the sequential program.
	 * 
	 * @param task is the input task value
	 * @return the result of the computation
	 */
	public Task[] compute(Task[] task) {
				
		return program.compute(task);
	}

	/**
	 * this is the method used to fetch the graph when compiling
	 * the class as a skeleton parameter.
	 * 
	 * @return the associated MDF graph implementing the wrapped code
	 * 			(may be parallel)
	 */
	public MdfGraph getMdfGraph() {
		return program;
	}
	
	public Object clone() {
		return new ParCompute((MdfGraph) program.clone());
	}
	
}
