package skeleton;

import muskel.mdf.Task;

/**
 * Compute con Side Effects, cioe' con la capacita' di generare
 * un nuovo task da calcolare.
 * La sottoclasse dovra' implementare il solo metodo compute, all'interno della quale
 * potra' impostare il task generato con il metodo setGeneratedTask.
 * <br>
 * E' cruciale che la generazione dei task sia condizionata dall'input del metodo compute,
 * altrimenti si creeranno dei loop che porteranno alla generazione continua di task.
 * <br>
 * <br>
 * Es. Nel corpo del metodo compute potremmo avere (posto x l'input):
 * <blockquote><pre>
 * if(x > y) setGeneratedTask(new Task[]{new Task(newTask)});
 * </pre></blockquote>
 * Se SECompute e' condiviso da piu' thread, sara' necessaria la sincronizzazione.
 * Nello specifico e' necessario rendere atomico il calcolo e il recupero dei task generati.
 * <br>
 * <br>
 * Es. Posto seComp, di tipo SECompute, il codice per il calcolo e il recupero dei task generati deve essere:
 * <blockquote><pre>
 * Task[] generatedTask = null;
 * Task[] result;
 * 
 * synchronized (seComp) {
 *	result = (Task[]) seComp.compute(tokenValues);
 *					
 *	Task[] gtTemp = seComp.getGeneratedTask();
 *					
 *	if(gtTemp != null) {
 *												
 *		generatedTask = new Task[gtTemp.length];
 *		//clono i task generati
 *		for(int i = 0; i< gtTemp.length; i++)
 *			generatedTask[i] = (Task) gtTemp[i].clone();
 *						
 *		//reimposto il valore per il prossimo utilizzo
 *		seComp.resetGeneratedTask();
 *	}
 * }
 * </pre></blockquote>
 * */
public abstract class SECompute extends Compute {

	private Task[] generatedTasks = null; 
	
	/**
	 * Restituisce il task generato dalla SECompute
	 * */
	public Task[] getGeneratedTask() {
		return generatedTasks;
	}
	
	/**
	 * Imposta il task specificato come task generato dalla SECompute.
	 * Va utilizzato nel metodo compute
	 * */
	protected void setGeneratedTask(Task[] tasks) {
		generatedTasks = tasks;
	}
	
	/**
	 * Reimposta i task generati dalla SECompute.
	 * */
	public void resetGeneratedTask(){
		generatedTasks = null;
	}
	
	@Override
	public abstract Task[] compute(Task[] task);
	
}
